/*
 * Copyright (C) 2024 Das Land Schleswig-Holstein vertreten durch den
 * Ministerpräsidenten des Landes Schleswig-Holstein
 * Staatskanzlei
 * Abteilung Digitalisierung und zentrales IT-Management der Landesregierung
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */
package de.ozgcloud.alfa.bescheid;

import static org.assertj.core.api.Assertions.*;
import static org.mockito.ArgumentMatchers.*;
import static org.mockito.Mockito.*;

import java.util.Comparator;
import java.util.Optional;
import java.util.stream.Stream;

import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.DisplayName;
import org.junit.jupiter.api.Nested;
import org.junit.jupiter.api.Test;
import org.junit.jupiter.params.ParameterizedTest;
import org.junit.jupiter.params.provider.ValueSource;
import org.mockito.InjectMocks;
import org.mockito.Mock;
import org.mockito.Spy;

import de.ozgcloud.alfa.vorgang.VorgangHeaderTestFactory;

class BescheidServiceTest {

	@Spy
	@InjectMocks
	private BescheidService service;
	@Mock
	private BescheidRemoteService remoteService;
	@Mock
	private BescheidByBeschiedenAmComparator comparator;

	@Nested
	class TestGetBescheidDraft {

		@Test
		void shouldCallRemoteService() {
			service.getBescheidDraft(VorgangHeaderTestFactory.ID);

			verify(remoteService).getBescheidDraft(VorgangHeaderTestFactory.ID);
		}

		@Test
		void shouldReturnResultFromRemoteService() {
			var remoteServiceResult = Optional.of(BescheidTestFactory.create());
			when(remoteService.getBescheidDraft(VorgangHeaderTestFactory.ID)).thenReturn(remoteServiceResult);

			var result = service.getBescheidDraft(VorgangHeaderTestFactory.ID);

			assertThat(result).isEqualTo(remoteServiceResult);
		}
	}

	@Nested
	class TestFindByVorgangId {
		private final Stream<Bescheid> bescheidsUnsorted = Stream.of(BescheidTestFactory.create());
		private final Stream<Bescheid> bescheidsSorted = Stream.of(BescheidTestFactory.create());

		@BeforeEach
		void setUpMocks() {
			when(remoteService.findByVorgangId(VorgangHeaderTestFactory.ID)).thenReturn(bescheidsUnsorted);
			doReturn(bescheidsSorted).when(service).sortBescheids(bescheidsUnsorted);
		}

		@Test
		void shouldCallRemoteService() {
			callService();

			verify(remoteService).findByVorgangId(VorgangHeaderTestFactory.ID);
		}

		@Test
		void shouldSortStream() {
			callService();

			verify(service).sortBescheids(bescheidsUnsorted);
		}

		@Test
		void shouldReturnSortedStream() {
			var resultStream = callService();

			assertThat(resultStream).isEqualTo(bescheidsSorted);
		}

		private Stream<Bescheid> callService() {
			return service.findByVorgangIdSorted(VorgangHeaderTestFactory.ID);
		}
	}

	@Nested
	class TestSortBescheids {
		@Mock
		private Stream<Bescheid> bescheids;

		@Mock
		private Comparator<Bescheid> reversedComparator;

		@BeforeEach
		void setUpMock() {
			when(comparator.reversed()).thenReturn(reversedComparator);
		}

		@Test
		void shouldSort() {
			service.sortBescheids(bescheids);

			verify(bescheids).sorted(reversedComparator);
		}

		@Test
		void shouldReturnSortedStream() {
			Stream<Bescheid> bescheidsSorted = Stream.of(BescheidTestFactory.create());
			when(bescheids.sorted(reversedComparator)).thenReturn(bescheidsSorted);

			var result = service.sortBescheids(bescheids);

			assertThat(result).isEqualTo(bescheidsSorted);
		}
	}

	@Nested
	class TestCanCreateBescheidDocumentAutomatically {

		@Test
		void shouldCallRemoteService() {
			service.canCreateBescheidDocumentAutomatically();

			verify(remoteService).canCreateBescheidDocument();
		}

		@ParameterizedTest
		@ValueSource(booleans = { true, false })
		void shouldRetrun(boolean canCreateBescheidDocument) {
			when(remoteService.canCreateBescheidDocument()).thenReturn(canCreateBescheidDocument);

			var canCreateAutomatically = service.canCreateBescheidDocumentAutomatically();

			assertThat(canCreateAutomatically).isEqualTo(canCreateBescheidDocument);
		}
	}

	@DisplayName("Exists bescheid")
	@Nested
	class TestExistsBescheid {

		@Test
		void shouldCallRemoteService() {
			service.existsBescheid(VorgangHeaderTestFactory.ID);

			verify(remoteService).findByVorgangId(VorgangHeaderTestFactory.ID);
		}

		@DisplayName("on existing bescheid")
		@Nested
		class TestOnExistingBescheid {

			@BeforeEach
			void mock() {
				when(remoteService.findByVorgangId(anyString())).thenReturn(Stream.of(BescheidTestFactory.create()));
			}

			@Test
			void shouldReturnTrue() {
				var exists = service.existsBescheid(VorgangHeaderTestFactory.ID);

				assertThat(exists).isTrue();
			}
		}

		@DisplayName("on missing bescheid")
		@Nested
		class TestOnMissingBescheid {

			@BeforeEach
			void mock() {
				when(remoteService.findByVorgangId(anyString())).thenReturn(Stream.empty());
			}

			@Test
			void shouldReturnFalse() {
				var exists = service.existsBescheid(VorgangHeaderTestFactory.ID);

				assertThat(exists).isFalse();
			}
		}
	}
}
