/*
 * Copyright (C) 2024 Das Land Schleswig-Holstein vertreten durch den
 * Ministerpräsidenten des Landes Schleswig-Holstein
 * Staatskanzlei
 * Abteilung Digitalisierung und zentrales IT-Management der Landesregierung
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */
package de.ozgcloud.alfa.common.zipdownload;

import static org.assertj.core.api.Assertions.*;
import static org.junit.jupiter.api.Assertions.*;
import static org.mockito.ArgumentMatchers.*;
import static org.mockito.Mockito.*;

import java.io.ByteArrayOutputStream;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.IOException;
import java.nio.file.Path;
import java.util.List;
import java.util.zip.ZipEntry;
import java.util.zip.ZipOutputStream;

import org.junit.jupiter.api.AfterEach;
import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.Nested;
import org.junit.jupiter.api.Test;
import org.mockito.InjectMocks;
import org.mockito.Mock;
import org.mockito.MockedConstruction;
import org.mockito.Spy;

import de.ozgcloud.alfa.common.binaryfile.BinaryFileService;
import de.ozgcloud.alfa.common.file.OzgFile;
import de.ozgcloud.alfa.common.file.OzgFileTestFactory;
import de.ozgcloud.common.binaryfile.TempFileUtils;
import de.ozgcloud.common.errorhandling.TechnicalException;
import lombok.SneakyThrows;

public class ZipDownloadServiceTest {

	@Spy
	@InjectMocks
	private ZipDownloadService service;

	@Mock
	private BinaryFileService binaryFileService;

	@Nested
	class TestWrite {
		private final ByteArrayOutputStream outputStream = new ByteArrayOutputStream();
		private final List<OzgFile> ozgFiles = List.of(OzgFileTestFactory.create());
		@Mock
		private File zipFile;

		@BeforeEach
		void init() {
			doReturn(zipFile).when(service).createZipFile(ozgFiles);
			doNothing().when(service).deleteFile(any());
		}

		@Test
		void shouldCreateZipFile() {
			doNothing().when(service).writeZipFileContent(zipFile, outputStream);

			callService();

			verify(service).createZipFile(ozgFiles);
		}

		@Test
		void shouldWriteZipFileContent() {
			doNothing().when(service).writeZipFileContent(zipFile, outputStream);

			callService();

			verify(service).writeZipFileContent(zipFile, outputStream);
		}

		@Test
		@SneakyThrows
		void shouldDeleteFile() {
			doNothing().when(service).writeZipFileContent(zipFile, outputStream);

			callService();

			verify(service).deleteFile(zipFile);
		}

		@Test
		void shouldDeleteFileAfterException() {
			doThrow(new RuntimeException()).when(service).writeZipFileContent(zipFile, outputStream);

			assertThrows(RuntimeException.class, this::callService);

			verify(service).deleteFile(zipFile);
		}

		private void callService() {
			service.write(ozgFiles, outputStream);
		}
	}

	@Nested
	class TestCreateZipFile {
		private final OzgFile file1 = OzgFileTestFactory.createWithUniqueId();
		private final OzgFile file2 = OzgFileTestFactory.createWithUniqueId();
		private final List<OzgFile> ozgFiles = List.of(file1, file2);

		private MockedConstruction<FileOutputStream> mockFileOutputStream;
		private MockedConstruction<ZipOutputStream> mockZipOutputStream;
		private File file;
		private FileOutputStream fileOutputStream;

		@BeforeEach
		void setUpMocks() {
			mockFileOutputStream = mockConstruction(FileOutputStream.class,
					(mock, context) -> {
						file = (File) context.arguments().get(0);
					});
			mockZipOutputStream = mockConstruction(ZipOutputStream.class,
					(mock, context) -> {
						fileOutputStream = (FileOutputStream) context.arguments().get(0);
					});
		}

		@AfterEach
		void cleanUp() {
			mockFileOutputStream.close();
			mockZipOutputStream.close();
		}

		@Test
		void shouldCreateZipOutputStream() {
			callService();

			assertThat(mockZipOutputStream.constructed()).hasSize(1);
			assertThat(fileOutputStream).isEqualTo(mockFileOutputStream.constructed().get(0));
		}

		@Test
		void shouldAddOzgFilesToZip() {
			callService();

			verify(service).addOzgFileToZip(file1, mockZipOutputStream.constructed().get(0));
			verify(service).addOzgFileToZip(file2, mockZipOutputStream.constructed().get(0));
		}

		@Test
		void shouldReturnFile() {
			var resultFile = callService();

			assertThat(resultFile).isEqualTo(file);
		}

		@Test
		void shouldThrowTechnicalException() {
			doNothing().when(service).deleteFile(any());

			doThrow(TechnicalException.class).when(service).addOzgFileToZip(any(), any());

			assertThrows(TechnicalException.class, this::callService);
		}

		@Test
		void shouldDeleteFileAfterException() {
			doThrow(TechnicalException.class).when(service).addOzgFileToZip(any(), any());
			doNothing().when(service).deleteFile(any());

			assertThrows(TechnicalException.class, this::callService);

			verify(service).deleteFile(file);
		}

		private File callService() {
			return service.createZipFile(ozgFiles);
		}
	}

	@Nested
	class TestAddOzgFileToZip {
		private final OzgFile ozgFile = OzgFileTestFactory.create();

		@Mock
		private ZipOutputStream zipOutputStream;

		private MockedConstruction<ZipEntry> mockZipEntry;
		private String entryName;

		@BeforeEach
		void setUpMocks() {
			mockZipEntry = mockConstruction(ZipEntry.class,
					(mock, context) -> {
						entryName = (String) context.arguments().get(0);
					});
		}

		@AfterEach
		void cleanUp() {
			mockZipEntry.close();
		}

		@Test
		void shouldCreateZipEntry() {
			callService();

			assertThat(mockZipEntry.constructed()).hasSize(1);
		}

		@Test
		void shouldSetZipEntryName() {
			callService();

			assertThat(entryName).isEqualTo(ozgFile.getName());
		}

		@Test
		@SneakyThrows
		void shouldPutNextEntry() {
			callService();

			verify(zipOutputStream).putNextEntry(mockZipEntry.constructed().get(0));
		}

		@Test
		@SneakyThrows
		void shouldWriteContent() {
			callService();

			verify(binaryFileService).writeFileContent(ozgFile.getId(), zipOutputStream);
		}

		@Test
		@SneakyThrows
		void shouldCloseEntry() {
			callService();

			verify(zipOutputStream).closeEntry();
		}

		@Test
		@SneakyThrows
		void shouldThrowTechnicalException() {
			doThrow(IOException.class).when(zipOutputStream).putNextEntry(any());

			assertThrows(TechnicalException.class, this::callService);
		}

		private void callService() {
			service.addOzgFileToZip(ozgFile, zipOutputStream);
		}
	}

	@Nested
	class TestWriteZipFileContent {
		private MockedConstruction<FileInputStream> mockFileInputStream;

		private File argumentFile;
		@Mock
		private File file;

		@Mock
		private FileOutputStream fileOutputStream;

		@AfterEach
		void cleanUp() {
			mockFileInputStream.close();
		}

		@Test
		void shouldCreateFileInputStreamWithFileArgument() {
			mockFileInputStream = mockConstruction(FileInputStream.class,
					(mock, context) -> {
						argumentFile = (File) context.arguments().get(0);
					});

			callService();

			assertThat(mockFileInputStream.constructed()).hasSize(1);
			assertThat(argumentFile).isEqualTo(file);
		}

		@Test
		@SneakyThrows
		void shouldTransferToOutputStream() {
			mockFileInputStream = mockConstruction(FileInputStream.class);

			callService();

			verify(mockFileInputStream.constructed().get(0)).transferTo(fileOutputStream);
		}

		@Test
		void shouldThrowTechnicalException() {
			mockFileInputStream = mockConstruction(FileInputStream.class,
					(mock, context) -> {
						when(mock.transferTo(fileOutputStream)).thenThrow(IOException.class);
					});

			assertThrows(TechnicalException.class, this::callService);
		}

		private void callService() {
			service.writeZipFileContent(file, fileOutputStream);
		}

	}

	@Nested
	class TestDeleteFile {

		@Test
		void shouldDeleteFile() {
			var file = TempFileUtils.createTmpFile().toFile();

			service.deleteFile(file);

			assertThat(file).doesNotExist();
		}

		@Test
		void shouldNotThrowException() {
			var file = Path.of("path").toFile();

			assertThatNoException().isThrownBy(() -> service.deleteFile(file));
		}

	}
}
