/*
 * Copyright (C) 2023 Das Land Schleswig-Holstein vertreten durch den
 * Ministerpräsidenten des Landes Schleswig-Holstein
 * Staatskanzlei
 * Abteilung Digitalisierung und zentrales IT-Management der Landesregierung
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */
package de.ozgcloud.alfa.postfach;

import static org.mockito.ArgumentMatchers.*;
import static org.mockito.Mockito.*;
import static org.springframework.test.web.servlet.request.MockMvcRequestBuilders.*;
import static org.springframework.test.web.servlet.result.MockMvcResultMatchers.*;

import java.io.IOException;
import java.io.OutputStream;
import java.time.LocalDate;
import java.util.stream.Stream;

import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.Disabled;
import org.junit.jupiter.api.DisplayName;
import org.junit.jupiter.api.Nested;
import org.junit.jupiter.api.Test;
import org.mockito.InjectMocks;
import org.mockito.Mock;
import org.mockito.Spy;
import org.springframework.http.HttpHeaders;
import org.springframework.http.MediaType;
import org.springframework.test.web.servlet.MockMvc;
import org.springframework.test.web.servlet.ResultActions;
import org.springframework.test.web.servlet.setup.MockMvcBuilders;
import org.springframework.web.servlet.mvc.method.annotation.StreamingResponseBody;

import de.ozgcloud.alfa.common.binaryfile.BinaryFileController;
import de.ozgcloud.alfa.vorgang.VorgangController;
import de.ozgcloud.alfa.vorgang.VorgangHeaderTestFactory;
import de.ozgcloud.alfa.vorgang.VorgangWithEingang;
import de.ozgcloud.alfa.vorgang.VorgangWithEingangTestFactory;
import lombok.SneakyThrows;

class PostfachMailControllerTest {

	@Spy
	@InjectMocks // NOSONAR
	private PostfachMailController controller;

	@Mock
	private PostfachMailService postfachMailService;

	@Mock
	private PostfachSettingsService postfachSettingsService;

	@Mock
	private PostfachMailModelAssembler assembler;

	@Mock
	private VorgangController vorgangController;
	@Mock
	private BinaryFileController binaryFileController;

	private MockMvc mockMvc;

	@BeforeEach
	void initTest() {
		mockMvc = MockMvcBuilders.standaloneSetup(controller).build();
	}

	@DisplayName("Get all")
	@Nested
	class TestGetAll {

		private final VorgangWithEingang vorgang = VorgangWithEingangTestFactory.create();
		private final Stream<PostfachMail> postfachMails = Stream.of(PostfachMailTestFactory.create());
		private final Stream<PostfachMail> sortedPostfachMails = Stream.of(PostfachMailTestFactory.create());
		private final PostfachSettings postfachSettings = PostfachSettingsTestFactory.create();

		@BeforeEach
		void mockVorgangController() {
			when(vorgangController.getVorgang(anyString())).thenReturn(vorgang);
			when(postfachMailService.getAll(VorgangHeaderTestFactory.ID)).thenReturn(postfachMails);
			doReturn(sortedPostfachMails).when(controller).sort(postfachMails);
			when(postfachSettingsService.getPostfachSettings(vorgang)).thenReturn(postfachSettings);
		}

		@Test
		void shouldCallService() {
			doRequest();

			verify(postfachMailService).getAll(VorgangHeaderTestFactory.ID);
		}

		@Test
		void shouldCallVorgangController() {
			doRequest();

			verify(vorgangController).getVorgang(VorgangHeaderTestFactory.ID);
		}

		@Test
		void shouldGetPostfachSettings() {
			doRequest();

			verify(postfachSettingsService).getPostfachSettings(vorgang);
		}

		@Test
		void shouldCallModelAssembler() {
			doRequest();

			verify(assembler).toCollectionModel(sortedPostfachMails, vorgang, postfachSettings);
		}

		@Test
		void shouldSortPostfachMails() {
			doRequest();

			verify(controller).sort(postfachMails);
		}

		@SneakyThrows
		private void doRequest() {
			mockMvc.perform(get(PostfachMailController.PATH + "?" + PostfachMailController.PARAM_VORGANG_ID + "=" + VorgangHeaderTestFactory.ID))
					.andExpect(status().isOk());
		}
	}

	@DisplayName("Get all as pdf")
	@Nested
	class TestGetAllAsPdf {

		@Mock
		private StreamingResponseBody streamingBody;

		private VorgangWithEingang vorgang = VorgangWithEingangTestFactory.create();

		@BeforeEach
		void mockService() {
			when(vorgangController.getVorgang(anyString())).thenReturn(vorgang);
		}

		@Test
		void shouldGetVorgang() {
			doRequest();

			verify(vorgangController).getVorgang(VorgangHeaderTestFactory.ID);
		}

		@Disabled("FIXME: Kippt um, wenn man alles Tests ausfuehrt")
		@Test
		void shouldCallService() {
			doRequest();

			verify(postfachMailService).getAllAsPdf(eq(vorgang), any(OutputStream.class));
		}

		@Test
		void shouldBuildResponseEntity() {
			doReturn(streamingBody).when(controller).createDownloadStreamingBody(vorgang);

			doRequest();

			verify(controller).buildResponseEntity(vorgang, streamingBody);
		}

		@SneakyThrows
		@Test
		void shouldReturnResponse() {
			doReturn(streamingBody).when(controller).createDownloadStreamingBody(vorgang);

			doRequest()
					.andExpect(header().string(HttpHeaders.CONTENT_DISPOSITION, buildPdfName()))
					.andExpect(content().contentType(MediaType.APPLICATION_PDF));
		}

		private String buildPdfName() {
			var today = LocalDate.now();
			var expectedDateFormat = String.valueOf(today.getYear()) + buildMonthValue() + buildDayValue();
			return VorgangHeaderTestFactory.NUMMER + "_" + expectedDateFormat + "_Nachrichten.pdf";
		}

		private String buildMonthValue() {
			return fillUpToTwoChars(LocalDate.now().getMonthValue());
		}

		private String buildDayValue() {
			return fillUpToTwoChars(LocalDate.now().getDayOfMonth());
		}

		private String fillUpToTwoChars(int value) {
			var asStr = String.valueOf(value);
			if (asStr.length() == 1) {
				return "0" + asStr;
			}
			return asStr;
		}

		@SneakyThrows
		private ResultActions doRequest() {
			return mockMvc
					.perform(get(PostfachMailController.PATH + "?" + PostfachMailController.PARAM_VORGANG_ID + "=" + VorgangHeaderTestFactory.ID)
							.accept(MediaType.APPLICATION_PDF_VALUE))
					.andExpect(status().isOk());
		}

	}

	@DisplayName("Create download streamingbody")
	@Nested
	class TestCreateDownloadStreamingBody {

		@Mock
		private OutputStream out;

		private VorgangWithEingang vorgang = VorgangWithEingangTestFactory.create();

		@Test
		void shouldCallServiceGetAllAsPdf() throws IOException {
			controller.createDownloadStreamingBody(vorgang).writeTo(out);

			verify(postfachMailService).getAllAsPdf(eq(vorgang), any());
		}

	}

	@DisplayName("Find postfach attachments")
	@Nested
	class TestFindPostfachAttachments {

		@BeforeEach
		void init() {
			when(postfachMailService.findById(any())).thenReturn(PostfachMailTestFactory.create());
		}

		@Test
		void shouldReturnSuccess() throws Exception {
			doRequest().andExpect(status().isOk());
		}

		@Test
		void shouldCallService() throws Exception {
			doRequest();

			verify(postfachMailService).findById(PostfachNachrichtId.from(PostfachMailTestFactory.ID));
		}

		@Test
		void shouldCallFileController() throws Exception {
			doRequest();

			verify(binaryFileController).getFiles(PostfachMailTestFactory.ATTACHMENTS);
		}

		private ResultActions doRequest() throws Exception {
			return mockMvc.perform(get(PostfachMailController.PATH + "/{nachrichtId}/attachments", PostfachMailTestFactory.ID));
		}

	}

	@DisplayName("Is postfach configured")
	@Nested
	class TestIsPostfachConfigured {

		@Test
		void shouldCallService() {
			controller.isPostfachConfigured();

			verify(postfachSettingsService).isPostfachConfigured();
		}

	}
}