/*
 * Copyright (C) 2024 Das Land Schleswig-Holstein vertreten durch den
 * Ministerpräsidenten des Landes Schleswig-Holstein
 * Staatskanzlei
 * Abteilung Digitalisierung und zentrales IT-Management der Landesregierung
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */
package de.ozgcloud.alfa.postfach;

import static org.assertj.core.api.Assertions.*;
import static org.mockito.Mockito.*;

import java.util.Collections;
import java.util.List;
import java.util.Map;
import java.util.Optional;
import java.util.UUID;

import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.DisplayName;
import org.junit.jupiter.api.Nested;
import org.junit.jupiter.api.Test;
import org.junit.jupiter.params.ParameterizedTest;
import org.junit.jupiter.params.provider.ValueSource;
import org.mockito.InjectMocks;
import org.mockito.Mock;
import org.mockito.Spy;
import org.springframework.cloud.endpoint.RefreshEndpoint;
import org.springframework.test.util.ReflectionTestUtils;

import com.thedeanda.lorem.LoremIpsum;

import de.ozgcloud.alfa.collaboration.OrganisationsEinheitTestFactory;
import de.ozgcloud.alfa.common.FeatureToggleProperties;
import de.ozgcloud.alfa.vorgang.ServiceKontoTestFactory;
import de.ozgcloud.alfa.vorgang.VorgangHeadTestFactory;
import de.ozgcloud.alfa.vorgang.VorgangWithEingang;
import de.ozgcloud.alfa.vorgang.VorgangWithEingangTestFactory;

class PostfachSettingsServiceTest {

	@Spy
	@InjectMocks
	private PostfachSettingsService service;
	@Mock
	private PostfachMailRemoteService remoteService;
	@Mock
	private RefreshEndpoint refreshEndpoint;
	@Mock
	private PostfachProperties postfachProperties;
	@Mock
	private FeatureToggleProperties featureToggleProperties;

	@DisplayName("Is postfach configured")
	@Nested
	class TestIsPostfachConfigured {

		@Test
		void shouldCallRemoteServiceOnMissingProperty() {
			when(remoteService.getPostfachConfig()).thenReturn(PostfachConfigGroupTestFactory.create());

			callService();

			verify(remoteService).getPostfachConfig();
		}

		@Test
		void shouldNotCallRemoteServiceOnExistingProperty() {
			setPostfachConfigGroup(PostfachConfigGroupTestFactory.create());

			callService();

			verifyNoInteractions(remoteService);
		}

		@Test
		void shouldCallRemoteServiceOnlyOnce() {
			when(remoteService.getPostfachConfig()).thenReturn(PostfachConfigGroupTestFactory.create());

			callService();
			callService();

			verify(remoteService, times(1)).getPostfachConfig();
		}

		@ParameterizedTest
		@ValueSource(booleans = { false, true })
		void shouldReturnValueFromRemoteService(boolean expectedConfigured) {
			var serviceResponse = PostfachConfigGroupTestFactory.createBuilder().configured(expectedConfigured).build();
			when(remoteService.getPostfachConfig()).thenReturn(serviceResponse);

			var configured = callService();

			assertThat(configured).isEqualTo(expectedConfigured);
		}

		@ParameterizedTest
		@ValueSource(booleans = { false, true })
		void shouldReturnCachedValue(boolean expectedConfigured) {
			setPostfachConfigGroup(PostfachConfigGroupTestFactory.createBuilder().configured(expectedConfigured).build());

			var configured = callService();

			assertThat(configured).isEqualTo(expectedConfigured);
		}

		private boolean callService() {
			return service.isPostfachConfigured();
		}
	}

	@Nested
	class TestGetPostfachSettings {
		private final VorgangWithEingang vorgang = VorgangWithEingangTestFactory.create();

		@BeforeEach
		void mockService() {
			doReturn(true).when(service).isReplyToMessageAllowed(vorgang);

		}

		@Test
		void shouldCheckIsReplyAllowed() {
			callService();

			verify(service).isReplyToMessageAllowed(vorgang);
		}

		@ParameterizedTest
		@ValueSource(booleans = { true, false })
		void shouldSetIsReplyAllowedFeature(boolean replyAllowed) {
			when(service.isReplyToMessageAllowed(vorgang)).thenReturn(replyAllowed);

			var postfach = callService();

			assertThat(postfach.getFeatures().isReply()).isEqualTo(replyAllowed);
		}

		@Test
		void shouldGetSignatur() {
			callService();

			verify(service).getSignatur(vorgang.getOrganisationseinheitenID());
		}

		@Test
		void shouldSetSignatur() {
			doReturn(SettingsTestFactory.SIGNATUR).when(service).getSignatur(vorgang.getOrganisationseinheitenID());

			var postfach = callService();

			assertThat(postfach.getSettings().getSignatur()).isEqualTo(SettingsTestFactory.SIGNATUR);
		}

		private PostfachSettings callService() {
			return service.getPostfachSettings(vorgang);
		}
	}

	@Nested
	class TestGetSignatur {

		@Test
		void shouldRefreshCloudProperties() {
			callService();

			verify(service).refreshPostfachProperties();
		}

		@Test
		void shouldGetOrganisationEinheitSettings() {
			callService();

			verify(service).getOrganisationsEinheitSettings(OrganisationsEinheitTestFactory.ID);
		}

		@Nested
		class OnExistingOrganisationsEinheitSignatur {

			@BeforeEach
			void setUp() {
				doReturn(Optional.of(OrganisationsEinheitSettingsTestFactory.create())).when(service)
						.getOrganisationsEinheitSettings(OrganisationsEinheitTestFactory.ID);
			}

			@Test
			void shouldReturnOrganisationsEinheitSignatur() {
				var signatur = callService();

				assertThat(signatur).isEqualTo(OrganisationsEinheitSettingsTestFactory.TEST_SIGNATUR);
			}

		}

		@Nested
		class OnNoOrganisationsEinheitSignatur {

			@BeforeEach
			void setUp() {
				doReturn(Optional.empty()).when(service)
						.getOrganisationsEinheitSettings(OrganisationsEinheitTestFactory.ID);
			}

			@Test
			void shouldGetSignatur() {
				callService();

				verify(postfachProperties).getSignatur();

			}

			@Test
			void shouldReturnSignatur() {
				when(postfachProperties.getSignatur()).thenReturn(SettingsTestFactory.SIGNATUR);

				var signatur = callService();

				assertThat(signatur).isEqualTo(SettingsTestFactory.SIGNATUR);
			}

		}

		private String callService() {
			return service.getSignatur(OrganisationsEinheitTestFactory.ID);
		}
	}

	@Nested
	class TestRefreshPostfachProperties {

		@Test
		void shouldCallRefreshEndpoint() {
			service.refreshPostfachProperties();

			verify(refreshEndpoint).refresh();
		}
	}

	@Nested
	@DisplayName("Is reply to a message allowed for given Vorgang")
	class TestIsReplyToMessageAllowed {

		@Test
		void shouldCallIsReplyAllowed() {
			doReturn(true).when(service).isReplyAllowed(ServiceKontoTestFactory.TYPE);

			service.isReplyToMessageAllowed(VorgangWithEingangTestFactory.create());

			verify(service).isReplyAllowed(ServiceKontoTestFactory.TYPE);
		}

		@ParameterizedTest
		@ValueSource(booleans = { false, true })
		void shouldReturnResultOfIsReplyAllowed(boolean expectedReplyAllowed) {
			doReturn(expectedReplyAllowed).when(service).isReplyAllowed(ServiceKontoTestFactory.TYPE);

			var replyAllowed = service.isReplyToMessageAllowed(VorgangWithEingangTestFactory.create());

			assertThat(replyAllowed).isEqualTo(expectedReplyAllowed);
		}

		@Test
		void shouldReturnFalseForNotExistingServiceKonto() {
			var vorgangWithEingang = VorgangWithEingangTestFactory.createBuilder()
					.header(VorgangHeadTestFactory.createBuilder()
							.serviceKonto(null)
							.build())
					.build();

			var replyToMessageAllowed = service.isReplyToMessageAllowed(vorgangWithEingang);

			assertThat(replyToMessageAllowed).isFalse();
		}
	}

	@Nested
	class TestIsReplyAllowed {

		private static final String DUMMY_SERVICE_KONTO_TYPE = LoremIpsum.getInstance().getWords(1);

		@BeforeEach
		void setUp() {
			when(featureToggleProperties.isReplyAlwaysAllowed()).thenReturn(false);
		}

		@Test
		@DisplayName("reply not allowed if postfach not configured")
		void shouldReturnFalseIfPostfachNotConfigured() {
			setPostfachConfigGroup(PostfachConfigGroupTestFactory.createBuilder().postfachConfigs(List.of()).build());
			doReturn(false).when(service).isPostfachConfigured();

			var replyAllowed = service.isReplyAllowed(PostfachConfigTestFactory.TYPE);

			assertThat(replyAllowed).isFalse();
		}

		@Test
		@DisplayName("given ServiceKonto-type is not configured")
		void shouldReturnFalseForNotConfiguredType() {
			setPostfachConfigGroup(PostfachConfigGroupTestFactory.createBuilder().postfachConfigs(List.of()).build());
			doReturn(true).when(service).isPostfachConfigured();

			var replyAllowed = service.isReplyAllowed(DUMMY_SERVICE_KONTO_TYPE);

			assertThat(replyAllowed).isFalse();
		}

		@Test
		@DisplayName("given ServiceKonto-type is null")
		void shouldReturnFalseForUnknownType() {
			setPostfachConfigGroup(PostfachConfigGroupTestFactory.create());
			doReturn(true).when(service).isPostfachConfigured();

			var replyAllowed = service.isReplyAllowed(null);

			assertThat(replyAllowed).isFalse();
		}

		@Test
		@DisplayName("reply not allowed for given ServiceKonto-type")
		void shouldReturnFalseIfReplyNotAllowed() {
			setPostfachConfigGroup(PostfachConfigGroupTestFactory.createBuilder()
					.postfachConfigs(List.of(PostfachConfigTestFactory.createBuilder().replyAllowed(false).build()))
					.build());
			doReturn(true).when(service).isPostfachConfigured();

			var replyAllowed = service.isReplyAllowed(PostfachConfigTestFactory.TYPE);

			assertThat(replyAllowed).isFalse();
		}

		@Test
		@DisplayName("reply allowed for given ServiceKonto-type")
		void shouldReturnTrueIfReplyIsAllowed() {
			setPostfachConfigGroup(PostfachConfigGroupTestFactory.create());
			doReturn(true).when(service).isPostfachConfigured();

			var replyAllowed = service.isReplyAllowed(PostfachConfigTestFactory.TYPE);

			assertThat(replyAllowed).isTrue();
		}

		@Test
		@DisplayName("reply allowed for given lower case ServiceKonto-type")
		void shouldReturnTrueIfReplyIsAllowedForLowerCaseServiceKontoType() {
			setPostfachConfigGroup(PostfachConfigGroupTestFactory.create());
			doReturn(true).when(service).isPostfachConfigured();

			var replyAllowed = service.isReplyAllowed(PostfachConfigTestFactory.TYPE.toLowerCase());

			assertThat(replyAllowed).isTrue();
		}

		@Test
		void shouldReturnTrueIfAlwaysAllowed() {
			when(featureToggleProperties.isReplyAlwaysAllowed()).thenReturn(true);

			var replyAllowed = service.isReplyAllowed(PostfachConfigTestFactory.TYPE);

			assertThat(replyAllowed).isTrue();
		}
	}

	private void setPostfachConfigGroup(PostfachConfigGroup postfachConfigGroup) {
		ReflectionTestUtils.setField(service, "postfachConfigGroup", postfachConfigGroup);
	}

	@Nested
	class TestGetOrganisationsEinheitSettings {

		@Nested
		class OnOrganisationIdNotNull {

			private final OrganisationsEinheitSettings organisationsEinheitSettings = OrganisationsEinheitSettingsTestFactory.create();
			private final Map<String, Object> organisationsEinheitSettingsMap = Map.of(PostfachSettingsService.FIELD_SIGNATUR,
					OrganisationsEinheitSettingsTestFactory.TEST_SIGNATUR);

			@BeforeEach
			void setUp() {
				when(postfachProperties.getOrganisationsEinheitSettings()).thenReturn(
						Map.of(OrganisationsEinheitTestFactory.ID, organisationsEinheitSettingsMap));
			}

			@Test
			void shouldGetOrganisationsEinheitSettings() {
				service.getOrganisationsEinheitSettings(OrganisationsEinheitTestFactory.ID);

				verify(postfachProperties).getOrganisationsEinheitSettings();
			}

			@Test
			void shouldMapToOrganisationsEinheitSettings() {
				service.getOrganisationsEinheitSettings(OrganisationsEinheitTestFactory.ID);

				verify(service).mapOrganisationsEinheitSettings(organisationsEinheitSettingsMap);
			}

			@Test
			void shouldReturnSettings() {
				var settings = service.getOrganisationsEinheitSettings(OrganisationsEinheitTestFactory.ID);

				assertThat(settings).isPresent().get().usingRecursiveComparison().isEqualTo(organisationsEinheitSettings);
			}

			@Test
			void shouldReturnEmpty() {
				var settings = service.getOrganisationsEinheitSettings(UUID.randomUUID().toString());

				assertThat(settings).isEmpty();
			}
		}

		@Nested
		class OnOrganisationIdNull {

			@Test
			void shouldReturnEmpty() {
				var settings = service.getOrganisationsEinheitSettings(null);

				assertThat(settings).isEmpty();
			}
		}
	}

	@Nested
	class TestMapOrganisationsEinheitSettings {

		private final String signatur = LoremIpsum.getInstance().getWords(2);
		private final Map<String, Object> organisationsEinheitSettings = Map.of(PostfachSettingsService.FIELD_SIGNATUR, signatur);

		@Test
		void shouldMap() {
			var settings = service.mapOrganisationsEinheitSettings(organisationsEinheitSettings);

			assertThat(settings.getSignatur()).isEqualTo(signatur);
		}

		@Test
		void shouldReturnSettingsWithNullSignatur() {
			var settings = service.mapOrganisationsEinheitSettings(Collections.emptyMap());

			assertThat(settings.getSignatur()).isNull();
		}
	}
}
